

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   DEVIL Screensaver System 1.05                                 *
    *   Copyright  2005 Sebastian Pipping <webmaster@hartwork.org>   *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2005-09-18   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*

	-- COMPILE INFO --
	libs:		MSVCRT.LIB kernel32.lib user32.lib gdi32.lib
	threads:	multi-threaded
	opts:		disable all default libs

*/



#include "Devil.h"
#include "DevilConfig.h"


#include <d3d8.h>
#include <d3dx8.h>
#include "Winamp/wa_ipc.h"
#include "Winamp/wa_msgids.h"
#include <list>

using namespace std;



#define D3DFVF_CUSTOM        ( D3DFVF_XYZ | D3DFVF_DIFFUSE )
#define FIRST_TIME_DOWN      0x40000000


// Compiletime config
#define MAX_NEWS_ENTRIES     20
#define NEWS_DISPLAY_TIME    1000
#define FPS_UPDATE_INTERVAL  1000
#define FONT_HEIGHT          24
#define GLOBE_RESOLUTION     13
#define DEFAULT_ALPHA        0xD0


#define GLOBE_TRIANGLES      ( GLOBE_RESOLUTION * 4 * ( GLOBE_RESOLUTION - 1 ) )



class Vertex
{
public:
	float fX;
	float fY;
	float fZ;
	DWORD dwColor;

	void Set( double fp_fX, double fp_fY, double fp_fZ, DWORD dwARGB )
	{
		fX =( float )fp_fX;
		fY =( float )fp_fY;
		fZ =( float )fp_fZ;
		dwColor = dwARGB;
	}
};



namespace News
{
	struct Item;

	list <Item *> Entries;
	
	struct Item
	{
		TCHAR * m_szText;
		DWORD m_dwBirth;
		DWORD m_dwDuration;
		
		Item( TCHAR * szText, DWORD dwMs )
		{
			const int iLen = _tcslen( szText );
			m_szText = ( TCHAR * )malloc( ( iLen + 1 ) * sizeof( TCHAR ) );
			memcpy( m_szText, szText, iLen * sizeof( TCHAR ) );
			m_szText[ iLen ] = TEXT( '\0' );
			m_dwBirth = GetTickCount();
			m_dwDuration = dwMs;
		}

		~Item()
		{
			free( m_szText );
		}
	};
	
	void Append( TCHAR * szText, DWORD dwDuration = NEWS_DISPLAY_TIME )
	{
		Item * new_item = new Item( szText, dwDuration );
		Entries.push_back( new_item );
		if( Entries.size() > MAX_NEWS_ENTRIES )
		{
			Entries.pop_front();
		}
	}
	
	void Update()
	{
		const DWORD dwNow = GetTickCount();
		list <Item *>::iterator iter = Entries.begin();
		while( iter != Entries.end() )
		{
			Item * cur_item = ( Item * )*iter;

			if( ( dwNow - cur_item->m_dwBirth ) > cur_item->m_dwDuration )
			{
				delete cur_item;
				Entries.pop_front();
				iter = Entries.begin();
			}
			else
			{
				break;
			}
		}
	}
}


D3DXMATRIX view_matrix;
IDirect3DVertexBuffer8 * g_Buffer = NULL;
IDirect3D8 * g_D3D = NULL;
IDirect3DDevice8 * g_Device = NULL;
ID3DXFont * g_Font = NULL;

Vertex g_Globe[ GLOBE_TRIANGLES * 3 ];


bool bBlendingEnabled = true;
bool bShowHelp = false;
bool bShowFPS = true;

float fAngle1 = 0.0f;
double fAngle2 = 0.0;

RECT FontRect = { 0, 0, 1024, 768 };

HWND hWinamp = NULL;



bool FACE_Live();
void FACE_Die();
LRESULT CALLBACK FACE_Master( HWND, UINT, WPARAM, LPARAM );
void FACE_Render( const double );
// void FACE_Config( const HWND );
bool FACE_Init3D();
void FACE_Cleanup3D();
// bool FACE_SwitchRes( int *, int *, UINT *, UINT * );
// void FACE_RestoreRes();



DEVIL_CONFIG g_CONFIG;

DEVIL_INTERFACE g_FACE =
{
	FACE_Live,				// Live
	FACE_Die,				// Die
	FACE_Master,			// Master
	FACE_Render,			// Render
	NULL,					// Config
	FACE_Init3D,			// Init3D
	FACE_Cleanup3D,			// Cleanup3D
	NULL,					// SwitchRes
	NULL,					// RestoreRes
	false,					// bEnableFrameLimiter
	60,						// uMaxFPS
	true,					// bMouseExit
	false,					// bAnyKeyExit

	NULL,					// hwnd
	DEVIL_MODE_UNKNOWN,		// Mode
	NULL,					// Slave
	0.0						// fFPS
};



// ***************************************************************************************
void BuildFont()
{
	LOGFONT log_Font = {
		FONT_HEIGHT,
		0,
		0,
		0,
		FW_NORMAL,
		FALSE,
		FALSE,
		FALSE,
		DEFAULT_CHARSET,
		OUT_TT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		PROOF_QUALITY,
		DEFAULT_PITCH | FF_MODERN,
		TEXT( "Courier New" )
	};

	D3DXCreateFontIndirect( g_Device, &log_Font, &g_Font );
}



// ***************************************************************************************
void DestroyFont()
{
   if( g_Font )
   {
      g_Font->Release();
      g_Font = NULL;
   }
}



// ***************************************************************************************
void RotationStep( double time_span )
{
	fAngle1 += 0.0005f * ( float )time_span;
	fAngle2 += 0.0001 * time_span;

	if( fAngle1 > D3DX_PI * 2 )
		fAngle1 -= D3DX_PI * 2;
	
	if( fAngle2 > D3DX_PI * 2 )
		fAngle2 -= D3DX_PI * 2;

	D3DXVECTOR3 vec( cos( fAngle2 ), 0, sin( fAngle2 ) );
	D3DXMatrixRotationAxis( &view_matrix, &vec, fAngle1 );
	g_Device->SetTransform( D3DTS_WORLD, &view_matrix );
}



// ***************************************************************************************
D3DCOLOR PositionToColor( int iAll, int iBrightness, int iColor )
{
	float fRed, fGreen, fBlue;
		float fSection = ( float )iColor / iAll / 2;
		fRed	= ( float )asin( sin( fSection * D3DX_PI ) ) * 2 / D3DX_PI;
		fGreen	= ( float )fabs( asin( sin( fSection * D3DX_PI + D3DX_PI / 3 ) ) * 2 / D3DX_PI );
		fBlue	= ( float )fabs( asin( sin( fSection * D3DX_PI + 2 * D3DX_PI / 3) ) * 2 / D3DX_PI );

	const int iMiddle = ( iAll - 1 ) / 2;
	const float fFactor = ( float )( iBrightness - iMiddle ) / iMiddle ;
	if( iBrightness > iMiddle )
	{
		// Brighten
		fRed   += ( 1.0f - fRed   ) * fFactor;
		fGreen += ( 1.0f - fGreen ) * fFactor;
		fBlue  += ( 1.0f - fBlue  ) * fFactor;
	}
	else
	{
		// Darken
		fRed += fRed * fFactor;
		fGreen += fGreen * fFactor;
		fBlue += fBlue * fFactor;
	}

	return D3DCOLOR_ARGB( DEFAULT_ALPHA, BYTE( 255 * fRed ), BYTE( 255 * fGreen ), BYTE( 255 * fBlue ) );
}



// ***************************************************************************************
void UpdateBlending()
{
	if( bBlendingEnabled )
	{
		g_Device->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
	 	g_Device->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
		g_Device->SetRenderState( D3DRS_BLENDOP, D3DBLENDOP_ADD );
		g_Device->SetRenderState( D3DRS_ALPHAFUNC, TRUE );
		g_Device->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
	}
	else
	{
		g_Device->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
	}
}



// ***************************************************************************************
bool FACE_Live()
{
	if( g_FACE.Mode == DEVIL_MODE_PREVIEW )
	{
		bBlendingEnabled = false;
		g_FACE.bEnableFrameLimiter = true;
		bShowFPS = false;
	}
	else
	{
		int res;
		g_CONFIG.Read( TEXT( "bBlendingEnabled" ), &res, 1 );
		bBlendingEnabled = ( bool )res;
		g_CONFIG.Read( TEXT( "bEnableFrameLimiter" ), &res, 0 );
		g_FACE.bEnableFrameLimiter = ( bool )res;
		g_CONFIG.Read( TEXT( "bShowFPS" ), &res, 1 );
		bShowFPS = ( bool )res;
	}
	return true;
}



// ***************************************************************************************
void FACE_Die()
{
	if( g_FACE.Mode == DEVIL_MODE_PREVIEW ) return;

	g_CONFIG.Write( TEXT( "bBlendingEnabled" ), ( int )bBlendingEnabled );
	g_CONFIG.Write( TEXT( "bEnableFrameLimiter" ), ( int )g_FACE.bEnableFrameLimiter );
	g_CONFIG.Write( TEXT( "bShowFPS" ), ( int )bShowFPS );
}



// ***************************************************************************************
LRESULT CALLBACK FACE_Master( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	static bool bInitDone = false;
	
	switch( message )
	{
	case WM_CREATE:
		if( !bInitDone )
		{
			if( g_FACE.Mode == DEVIL_MODE_PREVIEW )
			{
				bInitDone = true;
				break;
			}
			
			// Search running Winamp
			hWinamp = FindWindow( TEXT( "Winamp v1.x" ), NULL );
			if( hWinamp )
			{
				const int iWinampVersion = SendMessage( hWinamp, WM_WA_IPC, 0, IPC_GETVERSION );
				TCHAR szOut[ 255 ] = TEXT( "Running Winamp X.XX detected" );
				_itot( ( iWinampVersion & 0x0000f000 ) >> 12, szOut + 15, 16 );
				szOut[ 16 ] = TEXT( '.' );
				_itot( ( iWinampVersion & 0x000000f0 ) >>  4, szOut + 17, 16 );
				_itot( ( iWinampVersion & 0x0000000f )      , szOut + 18, 16 );
				szOut[ 19 ] = TEXT( ' ' );
				
				News::Append( szOut, 3500 );
			}
			else
			{
				News::Append( TEXT( "No running Winamp detected" ), 3500 );
			}
			
			News::Append( TEXT( "Press [H] for help" ), 4000 );

			bInitDone = true;	
		}
		break;
		
	case WM_KEYDOWN:
		if( g_FACE.Mode == DEVIL_MODE_PREVIEW ) break;

		switch( wp )
		{
		// Toggle help
		case 'H':
		case VK_F1:
			if( !( lp & FIRST_TIME_DOWN ) ) bShowHelp = !bShowHelp;
			break;

		// Toggle FPS
		case 'F':
		case VK_F2:
			if( !( lp & FIRST_TIME_DOWN ) ) bShowFPS = !bShowFPS;
			break;

		// Transparency
		case 'T':
		case VK_F3:
			if( !( lp & FIRST_TIME_DOWN ) )
			{
				bBlendingEnabled = !bBlendingEnabled;
				if( bBlendingEnabled )
					News::Append( TEXT( "Blending ON" ) );
				else
					News::Append( TEXT( "Blending OFF" ) );
				
				UpdateBlending();
			}
			break;

		// Toggle frame limiter
		case 'L':
		case VK_F4:
			if( !( lp & FIRST_TIME_DOWN ) )
			{
				g_FACE.bEnableFrameLimiter = !g_FACE.bEnableFrameLimiter;
				if( g_FACE.bEnableFrameLimiter )
					News::Append( TEXT( "Framelimiter ON" ) );
				else
					News::Append( TEXT( "Framelimiter OFF" ) );
			}
			break;

		// Play / Restart / Unpause
		case 'X':
		case VK_NUMPAD5:
			if( hWinamp && !( lp & FIRST_TIME_DOWN ) )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_BUTTON2, 0 );
			}
			break;

		// Stop / Stop with fadout / Stop after current track
		case 'V':
			if( hWinamp && !( lp & FIRST_TIME_DOWN ) )
			{
				const bool bShiftDown    = 128 & GetKeyState( VK_SHIFT );
				const bool bControlDown  = 128 & GetKeyState( VK_CONTROL );

				if( bShiftDown == bControlDown )
				{
					// Both or none
					PostMessage( hWinamp, WM_COMMAND, WINAMP_BUTTON4, 0 );
				}
				else
				{
					if( bShiftDown )
						PostMessage( hWinamp, WM_COMMAND, WINAMP_BUTTON4_SHIFT, 0 );
					else
						PostMessage( hWinamp, WM_COMMAND, WINAMP_BUTTON4_CTRL, 0 );
				}
			}
			break;

		// Rause / Unpause
		case 'C':
			if( hWinamp && !( lp & FIRST_TIME_DOWN ) )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_BUTTON3, 0 );
			}
			break;
		
		// Next track
		case 'B':
		case VK_NUMPAD6:
			if( hWinamp && !( lp & FIRST_TIME_DOWN ) )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_BUTTON5, 0 );
			}
			break;

		// Previous track
		case 'Z':
		case 'Y':			// For german keyboards
		case VK_NUMPAD4:
			if( hWinamp && !( lp & FIRST_TIME_DOWN ) )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_BUTTON1, 0 );
			}
			break;

		// Jump ten songs back
		case VK_NUMPAD1:
			if( hWinamp && !( lp & FIRST_TIME_DOWN ) )
			{
				PostMessage( hWinamp, WM_KEYDOWN, wp, lp );
			}
			break;

		// Jump ten songs forward
		case VK_NUMPAD3:
			if( hWinamp && !( lp & FIRST_TIME_DOWN ) )
			{
				PostMessage( hWinamp, WM_KEYDOWN, wp, lp );
			}
			break;

		// Rewind 5 seconds
		case VK_LEFT:
		case VK_NUMPAD7:
			if( hWinamp )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_REW5S, 0 );
			}
			break;

		// Fast-forward 5 seconds
		case VK_RIGHT:
		case VK_NUMPAD9:
			if( hWinamp )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_FFWD5S, 0 );
			}
			break;

		// Turn volume up
		case VK_UP:
		case VK_NUMPAD8:
			if( hWinamp )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_VOLUMEUP, 0 );
			}
			break;

		// Turn volume down
		case VK_DOWN:
		case VK_NUMPAD2:
			if( hWinamp )
			{
				PostMessage( hWinamp, WM_COMMAND, WINAMP_VOLUMEDOWN, 0 );
			}
			break;
		
		}
		break;
		
	}
	return CallWindowProc( g_FACE.Slave, hwnd, message, wp, lp );
}



// ***************************************************************************************
void FACE_Render( const double fPassedTime )
{
	News::Update();

	RotationStep( fPassedTime );
	
	
	// Clear the screen
	g_Device->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0 );

	// Enable Drawing
	g_Device->BeginScene();

	// Tell it what kind of vertices we're using
	g_Device->SetVertexShader( D3DFVF_CUSTOM );


	// Only back-faces
	g_Device->SetRenderState( D3DRS_CULLMODE, D3DCULL_CW );
	g_Device->SetStreamSource( 0, g_Buffer, sizeof( Vertex ) );
	g_Device->DrawPrimitive( D3DPT_TRIANGLELIST, 0, GLOBE_TRIANGLES );

	// Only front-faces
	g_Device->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
	g_Device->SetStreamSource( 0, g_Buffer, sizeof( Vertex ) );
	g_Device->DrawPrimitive( D3DPT_TRIANGLELIST, 0, GLOBE_TRIANGLES );


	if( bShowHelp || bShowFPS || !News::Entries.empty() )
	{
		const int iTopBackup = FontRect.top;
		
		g_Font->Begin();

		if( bShowHelp )
		{
			g_Font->DrawText(
				"H        F1       Toggle Help             \n"
				"F        F2       Toggle FPS              \n"
				"T        F3       Toggle Alpha Blending   \n"
				"L        F4       Toggle Frame Limiter    \n"
				"                                          \n"
				"                   -- Winamp keys --      \n"
				"X        Num 5    Play / Restart / Unpause\n"
				"V                 Stop                    \n"
				"Shift+V           Stop with fadeout       \n"
				"Ctrl+V            Stop after current track\n"
				"C                 Pause / Unpause         \n"
				"B        Num 6    Next track              \n"
				"Z Y      Num 4    Previous track          \n"
				"         Num 1    Jump 10 songs back      \n"
				"         Num 3    Jump 10 songs forward   \n"
				"Left     Num 7    Rewind 5 seconds        \n"
				"Right    Num 9    Fast-forward 5 seconds  \n"
				"Up       Num 8    Turn volume up          \n"
				"Down     Num 2    Turn volume down        ",
				-1,
				&FontRect,
				DT_RIGHT,
				0xffffffff
			);
		}

		if( bShowFPS )
		{
			static DWORD dwLastUpdate = 0;
			static float fLastFPS = 0.0f;

			const DWORD dwNow = GetTickCount();

			if( dwNow - dwLastUpdate > FPS_UPDATE_INTERVAL )
			{
				fLastFPS = g_FACE.fFPS;
				dwLastUpdate = dwNow;
			}
			
			static TCHAR szOut[ 12 ];
			_stprintf( szOut, TEXT( "FPS: %.02f" ), fLastFPS );

			g_Font->DrawText(
				szOut,
				-1,
				&FontRect,
				DT_LEFT,
				0xffffffff
			);
			
			FontRect.top += ( ( FONT_HEIGHT * 3 ) / 2 );
		}

		list<News::Item *>::iterator iter = News::Entries.begin();
		while( iter != News::Entries.end() )
		{
			g_Font->DrawText(
				( *iter )->m_szText,
				-1,
				&FontRect,
				DT_LEFT,
				0xffffffff
			);
			
			FontRect.top += FONT_HEIGHT;
			iter++;
		}

		g_Font->End();
		
		FontRect.top = iTopBackup;
	}


	// Disable drawing
	g_Device->EndScene();

	// Draw back buffer onto screen
	g_Device->Present( NULL, NULL, NULL, NULL );
}



/*
// ***************************************************************************************
void FACE_Config( const HWND hParent )
{
	// YOUR CODE HERE
}
*/


// ***************************************************************************************
bool FACE_Init3D()
{
	g_D3D = Direct3DCreate8( D3D_SDK_VERSION );
	if( !g_D3D )
	{
		return 0;
	}
	
	D3DDISPLAYMODE displayMode;
	HRESULT hr = g_D3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &displayMode );
	if( FAILED( hr ) )
	{
		return false;
	}


	int iWindowWidth;
	int iWindowHeight;
	RECT r;

	switch( g_FACE.Mode )
	{
	case DEVIL_MODE_FULLSCREEN:
		iWindowWidth   = displayMode.Width;
		iWindowHeight  = displayMode.Height;
		break;	

	case DEVIL_MODE_PREVIEW:
		GetClientRect( GetParent( g_FACE.hwnd ), &r );
		iWindowWidth   = r.right;
		iWindowHeight  = r.bottom;
		break;	
		
	default:
		GetClientRect( g_FACE.hwnd, &r );
		iWindowWidth   = r.right;
		iWindowHeight  = r.bottom;
	}


	// Font rect (relates to backbuffer, not window size!)
	const int ox      = iWindowWidth  / 50;
	const int oy      = iWindowHeight / 50;
	FontRect.left     = ox;
	FontRect.top      = oy;
	FontRect.right    = iWindowWidth - 2 * ox;
	FontRect.bottom   = iWindowHeight - 2 * oy;



	// Setup the present parameters
	D3DPRESENT_PARAMETERS presentParameters;
	memset( &presentParameters, 0, sizeof( D3DPRESENT_PARAMETERS ) );
	presentParameters.Windowed                = ( g_FACE.Mode != DEVIL_MODE_FULLSCREEN );
	presentParameters.SwapEffect              = D3DSWAPEFFECT_DISCARD;
	presentParameters.BackBufferFormat        = displayMode.Format;
	presentParameters.BackBufferWidth         = iWindowWidth; // displayMode.Width;
	presentParameters.BackBufferHeight        = iWindowHeight; // displayMode.Height;
	presentParameters.EnableAutoDepthStencil  = TRUE;
	presentParameters.AutoDepthStencilFormat  = D3DFMT_D16;


	hr = g_D3D->CreateDevice(
		D3DADAPTER_DEFAULT,
		D3DDEVTYPE_HAL,
		g_FACE.hwnd,
        D3DCREATE_SOFTWARE_VERTEXPROCESSING,
        &presentParameters,
		&g_Device
	); 

	if( FAILED( hr ) )
	{
		return false;
	}



	D3DXMATRIX mat;
	
	D3DXMatrixPerspectiveFovLH( &mat, D3DX_PI/10, ( float )iWindowWidth / iWindowHeight, 1.0, 100.0 );
	g_Device->SetTransform( D3DTS_PROJECTION, &mat );

	D3DXMatrixIdentity( &view_matrix );
	g_Device->SetTransform( D3DTS_WORLD, &view_matrix );

	D3DXMatrixTranslation( &mat, 0, 0, 9.0 );
	g_Device->SetTransform( D3DTS_VIEW, &mat );


	UpdateBlending();


	// Build globe
	ZeroMemory( g_Globe, GLOBE_TRIANGLES * sizeof( Vertex ) );
	int i = 0;
	for( int x = 0; x < GLOBE_RESOLUTION * 2; x++ )
	{
		const double dAngleX1 = x * D3DX_PI / GLOBE_RESOLUTION;
		const double dAngleX2 = ( x + 1 ) * D3DX_PI / GLOBE_RESOLUTION;

		const double fsinX1 = sin( dAngleX1 );
		const double fsinX2 = sin( dAngleX2 );
		const double fcosX1 = cos( dAngleX1 );
		const double fcosX2 = cos( dAngleX2 );

		// Top
		double dAngleY1 = 0;					// UPPER
		double dAngleY2 = D3DX_PI / GLOBE_RESOLUTION;		// LOWER

		double fsinY1 = 0;
		double fsinY2 = sin( dAngleY2 );
		double fcosY1 = 1;
		double fcosY2 = cos( dAngleY2 );

		D3DCOLOR color = PositionToColor( GLOBE_RESOLUTION, 0, x );

		g_Globe[ i     ].Set( -fcosX1 * fsinY1, fcosY1, fsinX1 * fsinY1, color ); // A
		g_Globe[ i + 1 ].Set( -fcosX1 * fsinY2, fcosY2, fsinX1 * fsinY2, color ); // B
		g_Globe[ i + 2 ].Set( -fcosX2 * fsinY2, fcosY2, fsinX2 * fsinY2, color ); // C
		
		i += 3;

		for( int y = 1; y < GLOBE_RESOLUTION - 1; y++ )
		{
			dAngleY1 = y * D3DX_PI / GLOBE_RESOLUTION;			// UPPER
			dAngleY2 = ( y + 1 ) * D3DX_PI / GLOBE_RESOLUTION;	// LOWER

			fsinY1 = sin( dAngleY1 );
			fsinY2 = sin( dAngleY2 );
			fcosY1 = cos( dAngleY1 );
			fcosY2 = cos( dAngleY2 );

			color = PositionToColor( GLOBE_RESOLUTION, y, x );

			// ABC
			g_Globe[ i     ].Set( -fcosX2 * fsinY1, fcosY1, fsinX2 * fsinY1, color ); // A
			g_Globe[ i + 1 ].Set( -fcosX1 * fsinY1, fcosY1, fsinX1 * fsinY1, color ); // B
			g_Globe[ i + 2 ].Set( -fcosX1 * fsinY2, fcosY2, fsinX1 * fsinY2, color ); // C
			
			// ACD
			g_Globe[ i + 3 ] = g_Globe[ i ];
			g_Globe[ i + 4 ] = g_Globe[ i + 2 ];
			g_Globe[ i + 5 ].Set( -fcosX2 * fsinY2, fcosY2, fsinX2 * fsinY2, color ); // D
			i += 6;
		}

		dAngleY1 = ( GLOBE_RESOLUTION - 1 ) * D3DX_PI / GLOBE_RESOLUTION;
		dAngleY2 = GLOBE_RESOLUTION * D3DX_PI / GLOBE_RESOLUTION;

		fsinY1 = sin( dAngleY1 );
		fsinY2 = sin( dAngleY2 );
		fcosY1 = cos( dAngleY1 );
		fcosY2 = cos( dAngleY2 );

		color = PositionToColor( GLOBE_RESOLUTION, GLOBE_RESOLUTION - 1, x );

		g_Globe[ i     ].Set( -fcosX2 * fsinY1, fcosY1, fsinX2 * fsinY1, color );
		g_Globe[ i + 1 ].Set( -fcosX1 * fsinY1, fcosY1, fsinX1 * fsinY1, color );
		g_Globe[ i + 2 ].Set( -fcosX1 * fsinY2, fcosY2, fsinX1 * fsinY2, color );

		i += 3;
	}
	
	
	// Depth testing OFF
	g_Device->SetRenderState( D3DRS_ZENABLE, FALSE );

	// Lightning OFF
	g_Device->SetRenderState( D3DRS_LIGHTING, FALSE );


	hr = g_Device->CreateVertexBuffer(
		GLOBE_TRIANGLES * 3 * sizeof( Vertex ),
		D3DUSAGE_WRITEONLY,
		D3DFVF_CUSTOM,
		D3DPOOL_MANAGED,
		&g_Buffer
	);

	if( FAILED( hr ) )
	{
		return false;
	}


	// Copy traingles to buffer
	BYTE * g_Buffer_content = NULL;
	g_Buffer->Lock( 0, 0, &g_Buffer_content, 0 );
	memcpy( g_Buffer_content, g_Globe, sizeof( Vertex ) * GLOBE_TRIANGLES * 3 );
	g_Buffer->Unlock();



	BuildFont();

	return true;
}



// ***************************************************************************************
void FACE_Cleanup3D()
{
	// Release device
	if( g_Device )
	{
        g_Device->Release();
		g_Device = NULL;
	}

	// Release Direct3D object
	if( g_D3D )
	{
		g_D3D->Release();
		g_D3D = NULL;
	}

	DestroyFont();

	// Release buffer
	if( g_Buffer )
	{
		g_Buffer->Release();
		g_Buffer = NULL;
	}
}


/*
// ***************************************************************************************
bool FACE_SwitchRes( int * x, int * y, UINT * uWidth, UINT * uHeight )
{
	// YOUR CODE HERE

	return true;
}



// ***************************************************************************************
void FACE_RestoreRes()
{
	// YOUR CODE HERE
}
*/
